#include "global.h"

#pragma config UPLLEN   = ON            // USB PLL Enabled
#pragma config FPLLMUL  = MUL_20        // PLL Multiplier
#pragma config UPLLIDIV = DIV_2         // USB PLL Input Divider
#pragma config FPLLIDIV = DIV_2         // PLL Input Divider
#pragma config FPLLODIV = DIV_1         // PLL Output Divider
#pragma config FPBDIV   = DIV_1         // Peripheral Clock divisor
#pragma config FWDTEN   = OFF           // Watchdog Timer
#pragma config WDTPS    = PS1           // Watchdog Timer Postscale
#pragma config FCKSM    = CSDCMD        // Clock Switching & Fail Safe Clock Monitor
#pragma config OSCIOFNC = OFF           // CLKO Enable
#pragma config POSCMOD  = HS            // Primary Oscillator
#pragma config IESO     = OFF           // Internal/External Switch-over
#pragma config FSOSCEN  = OFF           // Secondary Oscillator Enable (KLO was off)
#pragma config FNOSC    = PRIPLL        // Oscillator Selection
#pragma config CP       = OFF           // Code Protect
#pragma config BWP      = ON            // Boot Flash Write Protect
#pragma config PWP      = OFF           // Program Flash Write Protect
#pragma config ICESEL   = ICS_PGx1      // ICE/ICD Comm Channel Select
#pragma config DEBUG    = OFF           // Background Debugger Enable#else

FATFS	Fatfs;			/* File system object */

BOOL InitializeSystem ( void )
{
	SYSTEMConfigPerformance(GetSystemClock());
	INTEnableSystemMultiVectoredInt(); // Interrupt vectoring

	/* Start Timer1 in interval time of 1ms */
	OpenTimer1( T1_ON | T1_PS_1_256 | T1_SOURCE_INT,GetPeripheralClock()/256/1000);
	PR1 = GetPeripheralClock()/256/1000-1;
	mT1SetIntPriority(5);
	mT1ClearIntFlag();
	mT1IntEnable(1);

	AD1PCFG=0x1FFF;

	led1_setup();
	led2_setup();

	init_pullup();
	LATG=	0x0000;
	TRISG=	0xfcbf;		//RG6(SCK), RG8(SDO), RG9(CS) as output

    return TRUE;
}

void main(void)
{
	FIL		fil;
	FRESULT	fres;

	InitializeSystem();
	led1_off();
	led2_off();

	DelayMs(3000);

	f_mount(0, &Fatfs);
	fres=f_open(&fil,"HELLO.TXT",FA_WRITE | FA_CREATE_ALWAYS);

	if(fres==FR_OK){
		f_printf(&fil,"Hello World\n");
		f_close(&fil);
	}

	for(;;){
		led1_on();
		DelayMs(1000);
		led1_off();
		led2_on();
		DelayMs(1000);
		led2_off();
	}
}